from importlib import import_module
import inspect
from pathlib import Path
import sys
from types import ModuleType
from typing import Any, Callable, Union, cast, Optional
from typing_extensions import TypeAlias, TypeGuard

from nacolla.models import ImmutableModel
from nacolla.parsing.implementation_map_file_specification import Import
from nacolla.stateful_callable import StatefulCallable

ABSTRACT_STEP = Union[
    Callable[[ImmutableModel], ImmutableModel], type[StatefulCallable[Any, Any]]
]

IMPLEMENTATION: TypeAlias = Union[
    ABSTRACT_STEP,
    dict[str, ABSTRACT_STEP],
]


def parse_implementation(import_definition: Import) -> IMPLEMENTATION:
    module: ModuleType = _load_module(import_definition)

    if _is_specific_import(import_definition.name):
        abstract_step: ABSTRACT_STEP = _retrieve_function(
            module, import_definition.name
        )
        return abstract_step

    else:
        steps_dict: dict[str, ABSTRACT_STEP] = _retrieve_all_functions(module)
        return steps_dict


def _load_module(import_definition: Import) -> ModuleType:
    try:
        return import_module(import_definition.module)
    except (ModuleNotFoundError, TypeError):
        module_path: Path = Path(import_definition.module)
        if not module_path.exists():
            raise ValueError("Cannot find " + import_definition.module)

        sys.path.append(str(module_path.parents[0]))

        return import_module(name=module_path.stem, package=str(module_path.parents[0]))


def _retrieve_function(
    module: ModuleType, import_definition_name: str
) -> ABSTRACT_STEP:
    implementation = getattr(module, import_definition_name)
    if inspect.isclass(implementation):
        return _check_stateful_step(implementation, import_definition_name)
    else:
        return implementation


def _retrieve_all_functions(
    module: ModuleType,
) -> dict[str, ABSTRACT_STEP]:
    members = inspect.getmembers(module)
    public_functions: dict[str, Callable[[ImmutableModel], ImmutableModel]] = {
        member[0]: member[1]
        for member in members
        if inspect.isfunction(member[1]) and not member[0].startswith("_")
    }
    public_stateful_steps: dict[str, type[StatefulCallable[Any, Any]]] = {
        member[0]: member[1]
        for member in members
        if inspect.isclass(member[1])
        and issubclass(member[1], StatefulCallable)
        and member[1] != StatefulCallable
        and not member[0].startswith("_")
    }
    return {
        name: func
        for name, func in {**public_functions, **public_stateful_steps}.items()
    }


def _check_stateful_step(
    to_check: type[Any], import_definition_name: str
) -> type[StatefulCallable[Any, Any]]:
    if not issubclass(to_check, StatefulCallable):
        raise TypeError(
            "Implementation '"
            + str(import_definition_name)
            + "' is not a subclass of StatefulCallable."
        )
    return cast(type[StatefulCallable[Any, Any]], to_check)


def _is_specific_import(import_definition_name: Optional[str]) -> TypeGuard[str]:
    return bool(import_definition_name)



from __future__ import annotations
from typing import Generic, Iterator, TypeVar, Union, cast
from nacolla.models import ImmutableModel
from nacolla.step import End, Step
import copy

_SOURCE = TypeVar("_SOURCE", bound=ImmutableModel)


class Flow(Iterator[Step[ImmutableModel, ImmutableModel]], Generic[_SOURCE]):
    def __init__(self, root: Step[_SOURCE, ImmutableModel], source: _SOURCE) -> None:
        self.__current_step: Step[ImmutableModel, ImmutableModel] = root
        self.__current_message: ImmutableModel = source

    @property
    def current_message(self):
        return copy.deepcopy(self.__current_message)

    @property
    def current_step(self):
        return self.__current_step

    def __next__(self) -> Step[ImmutableModel, ImmutableModel]:
        self.__current_message = self.__current_step(self.__current_message)
        self.__current_step = self._stop(
            next(self.__current_step)[type(self.__current_message)]
        )
        return self.__current_step

    def __iter__(self) -> Iterator[Step[ImmutableModel, ImmutableModel]]:
        return self

    def _stop(
        self, step: Union[Step[ImmutableModel, ImmutableModel], End]
    ) -> Step[ImmutableModel, ImmutableModel]:
        if type(step) is End:
            raise StopIteration
        else:
            return cast(Step[ImmutableModel, ImmutableModel], step)



from __future__ import annotations
from typing import (
    Any,
    Union,
    Callable,
    get_args,
    get_origin,
    get_type_hints,
)


def unwrap_union(to_unwrap: type) -> set[type]:
    """Unwrap a union type in the set of composing types"""

    if get_origin(to_unwrap) is Union:
        return set(get_args(to_unwrap))
    else:
        return {to_unwrap}


def io_interface(function: Callable[[Any], Any]) -> tuple[set[type], set[type]]:
    """retrieve the io_interface of a function from its type hints"""

    type_hints: dict[str, type] = get_type_hints(function)

    if not type_hints.get("return"):
        raise TypeError("Step '" + str(function) + "' is missing output annotation")
    if len(type_hints) == 1:
        raise TypeError("Function '" + str(function) + "' is missing input annotation")
    if len(type_hints) > 2:
        raise TypeError(
            "Function '"
            + str(function)
            + "' has an invalid signature, too many parameters (exactly 1 required) '"
            + str(len(type_hints) - 1 - 1)
            + "'"
        )

    input_interface: set[type] = unwrap_union(list(type_hints.values())[0])

    output_interface: set[type] = unwrap_union(list((type_hints).values())[1])

    return input_interface, output_interface



from __future__ import annotations
from typing import TypeVar, Union, cast
from nacolla.models import ImmutableModel

from nacolla.step import Step
from functools import singledispatch

from nacolla.utilities import overlapping, register


_T = TypeVar("_T", bound=ImmutableModel)
_S = TypeVar("_S", bound=ImmutableModel)
_P = TypeVar("_P", bound=ImmutableModel)
_Q = TypeVar("_Q", bound=ImmutableModel)


def merge(s1: Step[_T, _S], s2: Step[_P, _Q]) -> Step[Union[_T, _P], Union[_S, _Q]]:
    """Merge two steps in one.
    The step resulting from the merge handles the union of the types that the two handle.
    Steps with overlapping interfaces cannot be merged.
    """

    def apply(input: Union[_T, _P]) -> Union[_S, _Q]:
        raise NotImplementedError

    if overlapping(s1, s2):
        raise TypeError(
            "Step '"
            + str(s1)
            + "' and '"
            + str(s2)
            + "' have overlapping input interfaces "
            + str(s1.input)
            + " and "
            + str(s2.input)
        )

    apply_dispatch = singledispatch(apply)
    register(apply_dispatch, s1, cast(set[type], s1.input))
    register(apply_dispatch, s2, cast(set[type], s2.input))

    return Step[Union[_T, _P], Union[_S, _Q]](
        input_interface=s1.input.union(s2.input),
        output_interface=s1.output.union(s2.output),
        apply=apply_dispatch,
        name=str(s1) + "_" + str(s2),
        next={
            **next(s1),
            **next(s2),
        },
    )



from __future__ import annotations
from pydantic import BaseModel
from pydantic.generics import GenericModel


class ImmutableModel(BaseModel):
    """Immutable Pydantic Model"""

    class Config:
        validate_all = True
        extra = "forbid"
        allow_mutation = False
        frozen = True
        use_enum_values = True
        validate_assignment = True
        arbitrary_types_allowed = True
        copy_on_model_validation = "deep"
        smart_union = True
        underscore_attrs_are_private = True


class StepModel(GenericModel):
    """Base model for steps"""

    class Config:
        validate_all = True
        extra = "forbid"
        use_enum_values = True
        validate_assignment = True
        arbitrary_types_allowed = True
        copy_on_model_validation = "deep"
        smart_union = True
        underscore_attrs_are_private = True



from typing import Dict, Union

from pydantic.types import StrictBool, StrictFloat, StrictInt, StrictStr
from pydantic import validator
from nacolla.models import ImmutableModel

STEP_INSTANTIATION_DESCRIPTION = Dict[
    str,
    Union[
        StrictStr,
        Dict[StrictStr, Union[StrictInt, StrictFloat, StrictBool, StrictStr]],
    ],
]


class StepInstantiation(ImmutableModel):
    description: Dict[str, STEP_INSTANTIATION_DESCRIPTION]

    @validator("description")
    def validate_description_structure(
        cls, description: STEP_INSTANTIATION_DESCRIPTION
    ):
        for key, value in description.items():
            if key == "init":
                if not type(value) is dict:
                    raise ValueError(
                        "Invalid 'init' found '"
                        + str(value)
                        + "' it must be a dictionary of keyword arguments to be passed to the step.\n"
                        + "Please do not use init as a type, it is a reserved name"
                    )
            else:
                if not type(value) is str:
                    raise ValueError(
                        "Invalid step found in mapping '"
                        + str(key)
                        + "' to '"
                        + str(value)
                        + "'. Each type must map to a step expressed as a string"
                    )


class FlowDescription(ImmutableModel):
    root: str
    steps: Dict[str, StepInstantiation]



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Set,
    TypeVar,
)
from nacolla.models import ImmutableModel

if TYPE_CHECKING:
    from nacolla.step import Step  # pragma: no cover


_I = TypeVar("_I", bound=ImmutableModel)
_O = TypeVar("_O", bound=ImmutableModel)
_II = TypeVar("_II", bound=ImmutableModel)
_OO = TypeVar("_OO", bound=ImmutableModel)


def overlapping(s1: Step[_I, _O], s2: Step[_II, _OO]) -> bool:
    """Check whether two steps have overlapping input interface"""
    return bool(s1.input.intersection(s2.input))


def register(
    registar: Callable[[Any], Any],  # must be a singledispatch callable
    to_register: Callable[[Any], Any],
    to_register_interface: Set[type],
) -> None:
    """Register callable to a singledispatch registar on a given type"""
    for type_to_register in to_register_interface:
        registar.register(type_to_register)(to_register)  # type: ignore



from pathlib import Path
from typing import Dict, Optional, Tuple, Type, TypeVar, Union
from typing_extensions import TypeGuard
from nacolla.models import ImmutableModel
from nacolla.parsing.parse_implementation_map import (
    IMPLEMENTATION_MAP,
)
from nacolla.step import Step
from nacolla.flow import Flow
import toml

_SOURCE = TypeVar("_SOURCE", bound=ImmutableModel)
_MAPPING = Dict[Type[ImmutableModel], Step[ImmutableModel, ImmutableModel]]
_FLOW_DESCRIPTION_ITEM = Union[str, Dict[str, str]]


def parse_flow(
    flow_description: Path, implementation_map: IMPLEMENTATION_MAP, source: _SOURCE
) -> Flow[_SOURCE]:
    root, mapping = _validate_flow_description(
        toml.loads(flow_description.read_text()), implementation_map
    )
    for step_name, step_mapping in mapping.items():
        step = implementation_map[step_name]

        for port, destination_step in step_mapping.items():
            step.concatenate(destination_step, port)

    return Flow(root=root, source=source)


def _validate_flow_description(
    flow_description: Dict[str, _FLOW_DESCRIPTION_ITEM],
    implementation_map: IMPLEMENTATION_MAP,
) -> Tuple[Step[ImmutableModel, ImmutableModel], Dict[str, _MAPPING],]:

    root = flow_description.get("root")
    if not _root_is_str(root):
        raise ValueError(
            "Invalid root provided "
            + str(root)
            + " please specify a root in the flow description file as the name of a step in the implementation file.\n"
            + "Root is a reserved name do not use it for any of your steps."
        )

    root_step: Step[ImmutableModel, ImmutableModel] = _validate_root(
        root_step_name=root,
        implementation_map=implementation_map,
        flow_description=flow_description,
    )
    mapping: Dict[str, _MAPPING] = {}

    for step_name, plain_mapping in {
        step_name: plain_mapping
        for step_name, plain_mapping in flow_description.items()
        if step_name != "root"
    }.items():
        step = implementation_map.get(step_name)
        if step is None:
            raise ValueError(
                "Could not find '" + str(step_name) + "' in implementation map."
            )

        step_ports: Dict[str, Type[ImmutableModel]] = {
            output_port.__name__: output_port for output_port in step.output
        }

        if not _is_mapping_dict(plain_mapping):
            raise ValueError(
                "Malformed mapping for "
                + str(step_name)
                + " got "
                + str(plain_mapping)
                + ".\n"
                + "Expected a mapping from type to step."
            )

        step_mapping: _MAPPING = {}
        for port_name, destination_step_name in plain_mapping.items():
            port = step_ports.get(port_name)
            if port is None:
                raise ValueError(
                    "Could not find '"
                    + str(port_name)
                    + "' among the types handled by '"
                    + str(step_name)
                )
            destination_step = implementation_map.get(destination_step_name)
            if destination_step is not None:
                step_mapping[port] = destination_step
        mapping[step.name] = step_mapping

    return root_step, mapping


def _is_mapping_dict(
    dict_to_validate: Union[str, Dict[str, str]]
) -> TypeGuard[Dict[str, str]]:
    if type(dict_to_validate) is not dict:
        return False

    for value in dict_to_validate.values():
        if type(value) is not str:
            return False
    return True


def _root_is_str(root: Optional[_FLOW_DESCRIPTION_ITEM]) -> TypeGuard[str]:
    return root is None or root is not str


def _validate_root(
    root_step_name: str,
    implementation_map: IMPLEMENTATION_MAP,
    flow_description: Dict[str, _FLOW_DESCRIPTION_ITEM],
) -> Step[ImmutableModel, ImmutableModel]:

    root: Step[ImmutableModel, ImmutableModel] = implementation_map[root_step_name]

    if root is None:
        raise ValueError(
            "Please specify a valid root in the description file, '"
            + str(root_step_name)
            + "' could not be found in the implementation file"
        )
    if root.name not in flow_description:
        raise ValueError(
            "Root step '"
            + str(root.name)
            + "' must be among the steps included in the flow '"
            + str(flow_description.keys())
            + "'."
        )
    return root



from types import MappingProxyType
from typing import (
    Any,
    Callable,
    Generic,
    Mapping,
    Optional,
    Union,
    cast,
)
from typing_extensions import TypeGuard, Self
from pydantic import validator, StrictStr
from nacolla.models import StepModel, ImmutableModel


from nacolla.utilities.generics import INPUT_INTERFACE, INTERFACE, OUTPUT_INTERFACE
from nacolla.utilities.type_utilities import io_interface


class End:
    def __next__(self):
        raise StopIteration


class Step(StepModel, Generic[INPUT_INTERFACE, OUTPUT_INTERFACE]):
    """A generic data transformation."""

    name: StrictStr
    apply: Callable[[INPUT_INTERFACE], OUTPUT_INTERFACE]

    # Optional at initialization, if not passed they get assembled
    # Access not None versions through input and output properties
    input_interface: Optional[set[type[INPUT_INTERFACE]]] = None
    output_interface: Optional[set[type[OUTPUT_INTERFACE]]] = None

    # further type specification makes pydantic recurse indefinitely
    # this field gets assembled after their interfaces if not user provided
    next: Optional[dict[type[OUTPUT_INTERFACE], Any]] = None

    def concatenate(
        self,
        destination_step: Self,
        port: type[INTERFACE],
    ) -> None:
        """Concatenate with a destination step along a given port"""
        # TODO: check that concatenation is possible by removing the cast and
        # adding a typeguard
        # to check if the port is among the outputs of this step
        # and the inputs of the destination step

        if self.next is None:
            raise ValueError("None next found, cannot concatenate")
        if self._is_concatenation_compatible(port, destination_step):
            self.next[port] = destination_step

    def __eq__(self, other: object) -> bool:
        """Compute equality between two steps.
        Steps with the same name are always equal"""

        if not isinstance(other, Step):
            return False
        return self.name == other.name

    def __call__(self, input: INPUT_INTERFACE) -> OUTPUT_INTERFACE:
        """Apply this step to a given input"""

        return self.apply(input)

    def __next__(
        self,
    ) -> MappingProxyType[type[OUTPUT_INTERFACE], Union[Self, End],]:
        """Retrieve all the next steps and their corresponding types.
        A read-only non-none view of the next steps is returned"""

        if self.next is not None:
            return cast(
                MappingProxyType[
                    type[OUTPUT_INTERFACE],
                    Union[Self, End],
                ],
                MappingProxyType(self.next),
            )
        raise ValueError("Found none mapping in " + str(self))

    def __str__(
        self,
    ):
        """A step is fully represented by its name"""
        return self.name

    def _is_concatenation_compatible(
        self, port: type, destination_step: Self
    ) -> TypeGuard[type[OUTPUT_INTERFACE]]:
        if (
            not next(self).get(cast(type[OUTPUT_INTERFACE], port))
            or port not in destination_step.input
        ):
            return False
        return True

    @property
    def input(self) -> set[type[INPUT_INTERFACE]]:
        """Input interface"""

        return cast(set[type[INPUT_INTERFACE]], self.input_interface)

    @property
    def output(self) -> set[type[OUTPUT_INTERFACE]]:
        """Output interface"""

        return cast(set[type[OUTPUT_INTERFACE]], self.output_interface)

    @validator("input_interface", "output_interface", pre=True, always=True)
    def assemble_interface(
        cls, interface: Optional[set[type]], values: dict[str, Any]
    ) -> set[type]:
        """Assemble i/o interface from apply's type hints"""

        if interface is not None:
            # interface set by user
            return interface

        apply = values["apply"]

        input_interface, output_interface = io_interface(apply)
        if not values.get("input_interface"):
            return input_interface
        if not values.get("output_interface"):
            return output_interface

        raise ValueError(
            "Input Interface and Output interface are already set, with values "
            + str(values.get("input_interface"))
            + " and "
            + str(values.get("output_interface"))
        )

    @validator("input_interface", "output_interface")
    def validate_interface(cls, interface: set[type]) -> set[type[ImmutableModel]]:
        """Validate i/o interface with regard to type hints"""

        if any(
            [
                not issubclass(interface_type, ImmutableModel)
                for interface_type in interface
            ]
        ):
            raise TypeError(
                "Input interface '"
                + "' must be either a subclass of ImmutableModel or a union of subclasses of ImmutableModel found "
                + str(interface)
            )
        return cast(set[type[ImmutableModel]], interface)

    @validator("next")
    def validate_next(
        cls,
        to_validate: Mapping[
            type[ImmutableModel],
            Union[
                Self,
                End,
            ],
        ],
        values: dict[str, Any],
    ) -> Mapping[type[ImmutableModel], Union[Self, End,],]:
        """Validate mapping to next steps with respect to interface compatibility"""

        if not values.get("output_interface"):
            raise ValueError("Interface parsing failed")

        if to_validate is None:
            return {output: End() for output in values["output_interface"]}

        Step._no_dangling_output_type(
            set(to_validate.keys()), values["output_interface"]
        )
        Step._no_incompatible_mapping(mappings=to_validate)
        return to_validate

    @staticmethod
    def _no_dangling_output_type(
        mapped_output_types: set[type], transformation_output_interface: set[type]
    ) -> None:
        """Check if the mapping contains all the output types of the step's transformation"""

        if not mapped_output_types == transformation_output_interface:
            raise TypeError(
                "Output types of step transformations are '"
                + str(mapped_output_types)
                + "'.\n While output types represented in the mapping to next steps are '"
                + str(transformation_output_interface)
                + "'. These two must be equal"
            )

    @staticmethod
    def _no_incompatible_mapping(
        mappings: Mapping[
            type[OUTPUT_INTERFACE],
            Union[
                Self,  # type: ignore (pyright signals an obscure error)
                End,
            ],
        ]
    ) -> None:
        """Check compatibility between input and output type in a mapping"""

        for in_type, step in mappings.items():
            if not isinstance(step, End):
                step = cast(Step[ImmutableModel, ImmutableModel], step)
                if in_type not in step.input:
                    raise TypeError(
                        "Output type '"
                        + str(in_type)
                        + "' is forwarded to '"
                        + str(step)
                        + "' which does not accept such input.\n It accepts only '"
                        + str(step.input)
                        + "'."
                    )



from typing import TypeVar

from nacolla.models import ImmutableModel


INTERFACE = TypeVar("INTERFACE", bound=ImmutableModel)
INPUT_INTERFACE = TypeVar("INPUT_INTERFACE", bound=ImmutableModel, contravariant=True)
OUTPUT_INTERFACE = TypeVar("OUTPUT_INTERFACE", bound=ImmutableModel, covariant=True)



from typing import Any, Optional, Union
from pydantic import validator
from pydantic.types import StrictStr

from nacolla.models import ImmutableModel

KWARG_TYPE = Union[int, float, bool, str]
KWARGS_SPECIFICATION = Union[dict[str, KWARG_TYPE], dict[str, dict[str, KWARG_TYPE]]]


class Import(ImmutableModel):
    module: StrictStr
    name: Optional[StrictStr] = ""


class ImplementationSpecification(ImmutableModel):
    callable: Import
    name: StrictStr
    kwargs: Optional[KWARGS_SPECIFICATION] = {}

    @property
    def arguments(self) -> KWARGS_SPECIFICATION:
        if self.kwargs is None:
            raise Exception("Found none kwargs for step " + self.name)
        return self.kwargs

    @validator("kwargs")
    def _kwargs_for_empty_name(
        cls, kwargs_to_validate: Optional[KWARGS_SPECIFICATION], values: dict[str, Any]
    ):
        if kwargs_to_validate is not None:
            if values["name"] == "":
                for key, value in kwargs_to_validate.items():
                    if not type(value) is dict:
                        raise ValueError(
                            "In case no name is specified you must specify kwargs as a dict from function/class name to kwargs.\n"
                            + "Error found at key '"
                            + str(key)
                            + "' with value '"
                            + str(value)
                            + "' in "
                            + str(values["callable"])
                        )
        return kwargs_to_validate


class ImplementationMapSpecification(ImmutableModel):
    implementations: list[ImplementationSpecification]

    @validator("implementations")
    def validate_implementations(
        cls, implementations: list[ImplementationSpecification]
    ) -> list[ImplementationSpecification]:
        names: list[str] = [implementation.name for implementation in implementations]
        if len(names) > len(set(names)):
            raise ValueError(
                "Found duplicate names in the implementation map file, this is forbidden"
            )
        return implementations



from __future__ import annotations
from functools import reduce
from pathlib import Path
from typing import Any, Callable, Dict, List, Type, Union, cast
import json
from typing_extensions import TypeGuard
from nacolla.models import ImmutableModel
from nacolla.operations import merge
from nacolla.parsing.implementation_map_file_specification import (
    KWARG_TYPE,
    KWARGS_SPECIFICATION,
    ImplementationMapSpecification,
)
from nacolla.parsing.parse_implementation import (
    ABSTRACT_STEP,
    IMPLEMENTATION,
    parse_implementation,
)
from nacolla.stateful_callable import StatefulCallable, make_stateful_step
from nacolla.step import Step
import inspect

IMPLEMENTATION_MAP = Dict[str, Step[ImmutableModel, ImmutableModel]]


def parse_implementation_map(
    implementation_map_file: Path,
) -> IMPLEMENTATION_MAP:
    specification: ImplementationMapSpecification = ImplementationMapSpecification(
        **json.loads(implementation_map_file.read_text())
    )

    implementation_dict: IMPLEMENTATION_MAP = {}
    for step in specification.implementations:

        parsed_implementation: IMPLEMENTATION = parse_implementation(
            import_definition=step.callable
        )
        if _to_merge(parsed_implementation):
            implementation_dict[step.name] = _parse_steps_and_merge(
                parsed_implementation, step.arguments, step.name
            )

        else:
            implementation_dict[step.name] = _parse_step(
                step.name,
                cast(
                    ABSTRACT_STEP,
                    parsed_implementation,
                ),
                kwargs=step.arguments,
            )

    return implementation_dict


def _parse_steps_and_merge(
    callable_dict: Dict[
        str,
        ABSTRACT_STEP,
    ],
    kwargs: KWARGS_SPECIFICATION,
    step_name: str,
):

    step_accumulator: List[Step[ImmutableModel, ImmutableModel]] = []
    for callable_name, callable_func in callable_dict.items():
        step_accumulator.append(
            _parse_step(callable_name, callable_func, kwargs=kwargs)
        )
    merged_step = reduce(merge, step_accumulator)  # type: ignore
    merged_step.name = step_name
    return merged_step


def _parse_step(
    callable_name: str,
    callable_func: ABSTRACT_STEP,
    kwargs: KWARGS_SPECIFICATION,
):
    parsed_kwargs: Dict[str, KWARG_TYPE] = cast(
        Dict[str, KWARG_TYPE],
        kwargs.get(callable_name)
        if isinstance(kwargs.get(callable_name), dict)
        else kwargs,
    )
    if _is_stateful_callable(callable_func):
        return make_stateful_step(callable_func(**parsed_kwargs), name=callable_name)

    if _is_callable(callable_func):
        return Step[ImmutableModel, ImmutableModel](
            name=callable_name, apply=callable_func
        )

    raise ValueError(
        "Could not build step from '"
        + str(callable_func)
        + "' got from '"
        + str(callable_name)
    )


def _is_stateful_callable(
    to_check: Any,
) -> TypeGuard[Type[StatefulCallable[ImmutableModel, ImmutableModel]]]:
    return inspect.isclass(to_check) and issubclass(to_check, StatefulCallable)


def _is_callable(
    to_check: Any,
) -> TypeGuard[Callable[[ImmutableModel], ImmutableModel]]:
    return callable(to_check)


def _to_merge(
    implementation: IMPLEMENTATION,
) -> TypeGuard[
    Dict[
        str,
        Union[
            Callable[[ImmutableModel], ImmutableModel],
            Type[StatefulCallable[Any, Any]],
        ],
    ],
]:
    return isinstance(implementation, dict)



from __future__ import annotations
from typing import Callable, Generic
import inspect
from nacolla.step import End, Step

from nacolla.utilities import io_interface
from nacolla.utilities.generics import INPUT_INTERFACE, OUTPUT_INTERFACE
from abc import ABC


class StatefulCallable(ABC, Generic[INPUT_INTERFACE, OUTPUT_INTERFACE]):
    def __init__(self) -> None:
        """Parse the i/o interface of the children from its public classes"""

        self.input_interface: set[type[INPUT_INTERFACE]] = set()
        self.output_interface: set[type[OUTPUT_INTERFACE]] = set()
        self.dispatch_dict: dict[
            type, Callable[[INPUT_INTERFACE], OUTPUT_INTERFACE]
        ] = {}

        public_methods: list[
            Callable[[INPUT_INTERFACE], OUTPUT_INTERFACE]
        ] = self._get_all_public_methods()

        for public_method in public_methods:
            input_interface, output_interface = StatefulCallable._parse_interface(
                public_method
            )
            self.input_interface |= input_interface
            self.output_interface |= output_interface

            self._update_dispatch_dict(input_interface, public_method)

    def __call__(self, input: INPUT_INTERFACE) -> OUTPUT_INTERFACE:
        """Route the input to the correct handler given its type and apply the transformation"""

        handler = self.dispatch_dict.get(type(input))
        if not handler:
            raise NotImplementedError("Cannot handle input of type " + str(type(input)))
        return handler(input)

    def _get_all_public_methods(
        self,
    ) -> list[Callable[[INPUT_INTERFACE], OUTPUT_INTERFACE]]:
        """Retrieve all the public methods of the children"""

        public_methods = [
            method
            for method_name, method in inspect.getmembers(
                self, predicate=inspect.ismethod
            )
            if not method_name.startswith("_") or method_name == "call"
        ]

        if not public_methods:
            raise TypeError(
                "No public methods defined in "
                + str(self.__class__)
                + " please define the interface of your steps through public methods"
            )
        return public_methods

    @staticmethod
    def _parse_interface(
        public_method: Callable[[INPUT_INTERFACE], OUTPUT_INTERFACE]
    ) -> tuple[set[type[INPUT_INTERFACE]], set[type[OUTPUT_INTERFACE]]]:
        """Parse the IO interface of a public method"""

        method_input_interface, method_output_interface = (
            Step.validate_interface(interface)
            for interface in io_interface(public_method)
        )
        return method_input_interface, method_output_interface

    def _update_dispatch_dict(
        self,
        input_interface: set[type[INPUT_INTERFACE]],
        public_method: Callable[[INPUT_INTERFACE], OUTPUT_INTERFACE],
    ):
        """Update the dispatching dictionary with a new input interface and method"""

        for interface_type in input_interface:
            self.dispatch_dict[interface_type] = public_method


def make_stateful_step(
    stateful_callable: StatefulCallable[INPUT_INTERFACE, OUTPUT_INTERFACE], name: str
) -> Step[INPUT_INTERFACE, OUTPUT_INTERFACE]:
    return Step[INPUT_INTERFACE, OUTPUT_INTERFACE](
        apply=stateful_callable,
        name=name,
        next={output: End() for output in stateful_callable.output_interface},
        input_interface=stateful_callable.input_interface,
        output_interface=stateful_callable.output_interface,
    )
